/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBM_SCCS) && !defined(lint)
/*static char sccsid[] = "from: @(#)announce.c	5.9 (Berkeley) 2/26/91";*/
static char rcsid[] = "$Id: announce.c,v 1.7 1998/03/24 23:57:21 dfaure Exp $";
#endif /* not lint */

/* Autoconf: */
#include <config.h>

#include "includ.h"
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <netdb.h>
#include <stdlib.h>
#include "print.h"
#include "announce.h"
#include "readconf.h"
#include "defs.h"

#ifdef HAVE_SGTTY_H
#include <sgtty.h>
#else
#ifdef HAVE_BSD_SGTTY_H
#include <bsd/sgtty.h>
#endif
#endif

#define max(a,b) ( (a) > (b) ? (a) : (b) )
#define N_LINES 5
#define N_CHARS 120

/*
 * Announce an invitation to talk.
 *
 * Because the tty driver insists on attaching a terminal-less
 * process to any terminal that it writes on, we must fork a child
 * to protect ourselves
 */
int announce(CTL_MSG *request, const char *remote_machine, char *disp) {

	int pid, val, status;

     if ((strstr(request->l_name,"\033"))||(strstr(remote_machine,"\033")))
	  {
	       if (strstr(remote_machine,"\033"))
		    {
			 syslog(LOG_WARNING, 
	      "blocked VT100 FLASH BOMB to user: %s (from field in packet contains bomb)", 
				request->r_name);
		    } else {
			 syslog(LOG_WARNING,
              "blocked VT100 FLASH BOMB to user: %s (apparently from: %s)", 
				request->r_name, remote_machine);
		    }
	  } else {
	    if ((pid = fork())) {
		/* we are the parent, so wait for the child */
		if (pid == -1)		/* the fork failed */
			return (FAILED);
		do {
			val = wait(&status);
			if (val == -1) {
				if (errno == EINTR)
					continue;
				/* shouldn't happen */
				syslog(LOG_WARNING, "announce: wait: %m");
				return (FAILED);
			}
		} while (val != pid);
		if ((status&0377) > 0)	/* we were killed by some signal */
			return (FAILED);
		/* Get the second byte, this is the exit/return code */
		return ((status >> 8) & 0377);
	    }
	    /* we are the child, go and do it */
	    _exit(announce_proc(request, remote_machine, disp));
	  }
     return (FAILED); /* should never happen */
}
	
/*
 * See if the user is accepting messages. If so, announce that 
 * a talk is requested.
 */

int announce_proc(CTL_MSG *request, const char *remote_machine, char *disp) {

  int usercfg = 0;    /* set if a user config file exists */
     
#ifdef HAVE_KDE

  int readPipe[2];
  char ch_aux;
  int pid, val, status;
  struct timeval clock;
  struct timezone zone;
  struct tm *localclock;
  char line_buf[N_CHARS];
  struct passwd * pw_buf;

  char * blankloc;

#ifdef PROC_FIND_USER
  if ((OPTXAnnounce) && (strlen(disp) >= 2)) {
#else
  if ((OPTXAnnounce) && ((int) request->r_tty[3] > (int) 'f')) {
#endif /* PROC_FIND_USER */

    /*
     * He is in X (probably :-) -> try to connect with external program
     */
      
      message_s("KDEDIR is %s",getenv("KDEDIR")); // just a check

    /* Open user config file. */
      usercfg = init_user_config(request->l_name);

    /*  No more needed : ktalkdlg will play sound itself.
        Other external programs can do whatever they want...
    sound_or_beep(usercfg); */

    gettimeofday(&clock, &zone);
    localclock = localtime( (const time_t *) &clock.tv_sec );
    (void)snprintf(line_buf, N_CHARS, "%s@%s", request->l_name,
                   remote_machine);
    
    /* disp can be a LIST of displays, such as ":0 :1", or ":0.0 :0.1"
        For the moment, let's take only the FIRST display in the list.
        To extend to ALL displays, we should do :
          wbegin=disp;
          end=disp+strlen(disp); // save because we'll insert zeros.
          for (p=disp; p<end, p++) {
            if (*p==' ') {
              *p='\0';
              ... current code, but use wbegin instead
               of disp...
                  pipe, switch(fork) {...}
              wbegin=p+1;
            }
          }

          But the problem is to know if ALL failed, or if at least one 
          succeeded.
          Other problem is the pipe. One or several ones ?
          
    */
    if ((blankloc=strstr(disp," "))) /* replace the first ' ' */
        *blankloc='\0';             /* by a \0 */

    pipe( readPipe );
    switch (pid = fork()) {
    case -1: {
        syslog(LOG_ERR,"Announce : Fork failed ! - %s",strerror(errno));
        return( FAILED );
    }
    case 0: {
        char extprg [S_MESSG]; /* Program used for notification. */

        if ((!usercfg) || (!read_user_config("ExtPrg", extprg, S_MESSG)))
            /* try to read extprg from user config file, otherwise : */
            strcpy(extprg,OPTextprg); /* take default */

        message_s("extprg is %s",extprg);
            
	/*
	 * set DISPLAY if it is not set yet (it is not) 
	 */

        setenv("DISPLAY", disp, 0);
        syslog(LOG_WARNING, "Trying to run %s...%s", extprg, getenv("DISPLAY"));
	/*
	 * point stdout, stderr of external program to the pipe
	 * announce a talk request by execing external program
	 */
	close( 1 );
	dup( readPipe[1] );
	close( 2 );
	dup( readPipe[1] );
        /* ===============================================
         * If the display uses Xauth, we will not be able to connect...
         * We try by changing the user ID...
         */
        pw_buf = getpwnam(request->r_name);
        if (setgid(pw_buf -> pw_gid) || setuid(pw_buf -> pw_uid))
            syslog(LOG_WARNING, "Warning: setuid: %m");
        else
            setenv("HOME", pw_buf -> pw_dir, 1);
        /*
         * End
         * ===============================================
         */

	execl( extprg, extprg, line_buf, 0 );

	/*
	 * failure - tell it to father
	 */
	ch_aux = '!';
	write( readPipe[1], &ch_aux, 1 );
	syslog(LOG_WARNING, "exec: %m");
	_exit( print_std_mesg( request, remote_machine, usercfg ) );
      }
      default: {
	/*
	 * I don't know any way how to find that external program run ok
         * so parent returns always SUCCESS
	 * Well I know a solution - read the pipe :-) - It blocks, so
         * we can't wait the reaction of the user
	 */
	read( readPipe[0], &ch_aux, 1 );
	syslog(LOG_WARNING, "Child process sent : %c",ch_aux); 
	close( readPipe[0] );
	close( readPipe[1] );
	if (ch_aux == '#') {
	  message("OK - Child process responded"); /* for debugging */
	  return( SUCCESS );
	} else {
	  message("KO - External program failed - normal announce");
	  do {
	    val = waitpid( pid, &status, 0 );
	    if (val == -1) {
	      if (errno == EINTR)
		continue;
	      /* shouldn't happen */
	      syslog(LOG_WARNING, "announce: wait: %m");
	      return (FAILED);
	    }
	  } while (val != pid);
	  if ((status&0377) > 0)	/* we were killed by some signal */
	    return (FAILED);
	  /* Get the second byte, this is the exit/return code */
	  return ((status >> 8) & 0377);
	}
      } /* default */
    } /* switch */
  } else {
#else
 {
#endif /* HAVE_KDE */
    /*
     * He is not in X -> print standard message
     */
    return( print_std_mesg( request, remote_machine, usercfg ) );
  }
 if (!usercfg) end_user_config(); 
}

int print_std_mesg( CTL_MSG *request, const char *remote_machine, int usercfg ) {

  char full_tty[32];
  FILE *tf;
  struct stat stbuf;
  
  (void)snprintf(full_tty, sizeof(full_tty), "%s/%s", _PATH_DEV, request->r_tty);
  if (access(full_tty, 0) != 0)
    return (FAILED);
  if ((tf = fopen(full_tty, "w")) == 0)
    return (PERMISSION_DENIED);
  /*
   * On first tty open, the server will have
   * it's pgrp set, so disconnect us from the
   * tty before we catch a signal.
   */
  ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
  if (fstat(fileno(tf), &stbuf) < 0)
    return (PERMISSION_DENIED);
  if ((stbuf.st_mode&020) == 0)
    return (PERMISSION_DENIED);
  print_mesg(tf, request, remote_machine, usercfg);
  fclose(tf);
  return (SUCCESS);
}

/*
 * Build a block of characters containing the message. 
 * It is sent blank filled and in a single block to
 * try to keep the message in one piece if the recipient
 * in in vi at the time
 */
void print_mesg(FILE * tf, CTL_MSG * request, const char *
                remote_machine, int usercfg)
{
	struct timeval clock;
	struct timezone zone;
	struct tm *localclock;
	char line_buf[N_LINES][N_CHARS];
	char buffer [N_CHARS];
	int sizes[N_LINES];
	char big_buf[N_LINES*N_CHARS];
	char *bptr, *lptr;
	int i, j, max_size;

	char *localname, *remotemach;
	const char *remotename;

	remotemach = strdup(remote_machine);
	/* We have to duplicate it because param_remote_machine is contained
	   in the hostent structure, and will be erased by gethostbyname. */
	
	localname = strdup(gethostbyname(hostname)->h_name);
	/* We have to duplicate localname also. Same reasons as above ! */
	
	remotename = gethostbyname(remotemach)->h_name;
	
	i = 0;
	max_size = 0;
	gettimeofday(&clock, &zone);
	localclock = localtime( (const time_t *) &clock.tv_sec );
	(void)snprintf(line_buf[i], N_CHARS, " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)snprintf(line_buf[i], N_CHARS, OPTannounce1,
		      localclock->tm_hour , localclock->tm_min );
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	snprintf(buffer, N_CHARS, "%s@%s", request->l_name, remote_machine);
	snprintf(line_buf[i], N_CHARS, OPTannounce2, buffer);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;

	if (!(strcmp(localname,remotename))) {
	  snprintf(line_buf[i], N_CHARS, OPTannounce3, request->l_name);
	} else {
	  snprintf(line_buf[i], N_CHARS, OPTannounce3, buffer);
	}

	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	(void)snprintf(line_buf[i], N_CHARS, " ");
	sizes[i] = strlen(line_buf[i]);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	bptr = big_buf;
	if (sound_or_beep(usercfg)) /* if no sound then : */
	  *bptr++ = ''; /* send something to wake them up */
	*bptr++ = '\r';	/* add a \r in case of raw mode */
	*bptr++ = '\n';
	for (i = 0; i < N_LINES; i++) {
		/* copy the line into the big buffer */
		lptr = line_buf[i];
		while (*lptr != '\0')
			*(bptr++) = *(lptr++);
		/* pad out the rest of the lines with blanks */
		for (j = sizes[i]; j < max_size + 2; j++)
			*(bptr++) = ' ';
		*(bptr++) = '\r';	/* add a \r in case of raw mode */
		*(bptr++) = '\n';
	}
	*bptr = '\0';
	fprintf(tf, big_buf);
	fflush(tf);
	ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
}

int play_sound(int usercfg)
{
     int pid, val, status, returncode;
     char sSoundPlayer[S_CFGLINE], sSoundFile[S_CFGLINE];
     char sSoundPlayerOpt[S_CFGLINE];
     /* We must absolutely read the configuration before forking,
        because the father will close the file soon !! */
     if ((!usercfg) || (!read_user_config("SoundPlayer",sSoundPlayer,S_CFGLINE)))
         strcpy(sSoundPlayer,OPTsoundplayer);
     message(sSoundPlayer);
     if ((!usercfg) || (!read_user_config("SoundPlayerOpt",sSoundPlayerOpt,S_CFGLINE)))
         strcpy(sSoundPlayerOpt,OPTsoundplayeropt);
     message(sSoundPlayerOpt);
     if ((!usercfg) || (!read_user_config("SoundFile",sSoundFile,S_CFGLINE)))
         strcpy(sSoundFile,OPTsoundfile);
     message(sSoundFile);

     if ((pid = fork())) {
	  /* we are the parent, so wait for the child */
	  if (pid == -1)          /* the fork failed */
	       syslog(LOG_ERR,"Fork before play_sound : FAILED.");
	       return (FAILED);
	  do {
	       val = wait(&status);
	       if (val == -1) {
		    if (errno == EINTR)
			 continue;
		    /* shouldn't happen */
		    syslog(LOG_WARNING, "announce: wait: %m");
		    return (FAILED);
	       }
	  } while (val != pid);
	  if ((status&0377) > 0)    /* we were killed by some signal */
	       return (FAILED);
	  /* Get the second byte, this is the exit/return code */
	  return ((status >> 8) & 0377);
     }
     /* we are the child, go and do it */
#if 0
     syslog(LOG_DEBUG, "%s %s %s %s",sSoundPlayer,sSoundPlayer/*arg0*/,sSoundFile,sSoundPlayerOpt);
#endif

     returncode = execl(sSoundPlayer,sSoundPlayer/*arg0*/,sSoundFile,sSoundPlayerOpt,NULL);

     message(strerror(errno));
     syslog(LOG_ERR,"ERROR PLAYING SOUND FILE !!!");
     syslog(LOG_ERR,"RETURN-CODE : %d",returncode);
     
     _exit(returncode);
}

/*
 *  Calls play_sound if necessary. Returns 1 if a tty beep is needed.
 */
int sound_or_beep(int usercfg)
{
     int bSound;
     int ret;
     if ((!usercfg) || (!read_bool_user_config("Sound",&bSound)))
         bSound=OPTsound;
     
     message2("Sound option in sound_or_beep : %d",bSound);
     
     if (bSound)
       {
	/* try to play sound, otherwise beeps (if not in X) */
	  ret = play_sound(usercfg); /* 1 = it didn't work. ^G needed */
       }
     else ret = 1; /* ^G needed */
     return ret;
}
